from scipy.stats import multinomial
from typing import List
from scipy.stats import norm
import tensorflow as tf
import numpy as np

import unittest
from easysrv import SRV

import os

os.environ["CUDA_DEVICE_ORDER"] = "PCI_BUS_ID"  # see issue #152
os.environ["CUDA_VISIBLE_DEVICES"] = ""


class SampleMSM:
    def __init__(self):
        self.p_init = np.array([0, 1, 0])
        self.p_transition = np.array(
            [
                [0.90, 0.05, 0.02, 0.03],
                [0.01, 0.90, 0.07, 0.02],
                [0.05, 0.03, 0.9, 0.02],
                [0.10, 0.03, 0.02, 0.85],
            ]
        )

    def markov_sequence(self, sequence_length=10000):
        """
        Generate a Markov sequence based on p_init and p_transition.
        """
        p_init = self.p_init
        p_transition = self.p_transition

        if p_init is None:
            p_init = equilibrium_distribution(p_transition)
        initial_state = list(multinomial.rvs(1, p_init)).index(1)

        states = [initial_state]
        for _ in range(sequence_length - 1):
            p_tr = p_transition[states[-1]]
            new_state = list(multinomial.rvs(1, p_tr)).index(1)
            states.append(new_state)
        return states

    def gaussian_emissions(self, states, mus, sigmas):
        emissions = []
        for state in states:
            loc = mus[state]
            scale = sigmas[state]
            e = norm.rvs(loc=loc, scale=scale)
            emissions.append(e)
        return np.array(emissions)

    def sample(self, n=10000, dims=2):
        dims = dims
        sep = 50
        mus = np.random.randn(4, dims) * sep
        sigmas = np.abs(np.random.randn(4, dims)) * sep
        gaussian_ems = self.gaussian_emissions(
            self.markov_sequence(sequence_length=n), mus=mus, sigmas=sigmas
        )
        return gaussian_ems


msm = SampleMSM()
sample_n = 10000
num = 40
data = [msm.sample(n=sample_n, dims=10) for x in range(num)]
model = tf.keras.models.Sequential(
    [
        tf.keras.layers.Dense(
            16,
            kernel_regularizer=tf.keras.regularizers.L2(0.0001),
            kernel_initializer="lecun_normal",
        ),
        tf.keras.layers.Activation("swish"),
        tf.keras.layers.BatchNormalization(),
        tf.keras.layers.Dense(
            16,
            kernel_regularizer=tf.keras.regularizers.L2(0.0001),
            kernel_initializer="lecun_normal",
        ),
        tf.keras.layers.Activation("swish"),
        tf.keras.layers.BatchNormalization(),
        tf.keras.layers.Dense(5, kernel_regularizer=tf.keras.regularizers.L2(0.0001)),
    ]
)
model(data[0])
optimizer = tf.keras.optimizers.Adam(learning_rate=0.001)
srv = SRV(model=model, optimizer=optimizer, epochs=5, lagtime=1, split=0.1)
srv.fit(data, verbose=False)


class TestFitting(unittest.TestCase):
    def test_remove_one_mode(self):
        out = srv.transform(data, remove_modes=[0])
        if out[0].shape[0] != sample_n:
            raise Exception(
                "Sample output dimension after trasform is different than assumed in the test"
            )
        if out[0].shape[-1] != 5 - 1:
            raise Exception(
                "Last output dimension after transform is different than assumed in the test"
            )
        if len(out) != num:
            raise Exception(
                "Number of samples after transform is different than assumed in the test"
            )

    def test_remove_two_subsequent_modes(self):
        out = srv.transform(data, remove_modes=[0, 1])
        if out[0].shape[0] != sample_n:
            raise Exception(
                "Sample output dimension after trasform is different than assumed in the test"
            )
        if out[0].shape[-1] != 5 - 2:
            raise Exception(
                "Last output dimension after transform is different than assumed in the test"
            )
        if len(out) != num:
            raise Exception(
                "Number of samples after transform is different than assumed in the test"
            )

    def test_remove_two_notsubsequent_modes(self):
        out = srv.transform(data, remove_modes=[1, 3])
        if out[0].shape[0] != sample_n:
            raise Exception(
                "Sample output dimension after trasform is different than assumed in the test"
            )
        if out[0].shape[-1] != 5 - 2:
            raise Exception(
                "Last output dimension after transform is different than assumed in the test"
            )
        if len(out) != num:
            raise Exception(
                "Number of samples after transform is different than assumed in the test"
            )

    def test_remove_all_except_last_mode(self):
        out = srv.transform(data, remove_modes=[0, 1, 2, 3])
        if out[0].shape[0] != sample_n:
            raise Exception(
                "Sample output dimension after trasform is different than assumed in the test"
            )
        if out[0].shape[-1] != 5 - 4:
            raise Exception(
                "Last output dimension after transform is different than assumed in the test"
            )
        if len(out) != num:
            raise Exception(
                "Number of samples after transform is different than assumed in the test"
            )


if __name__ == "__main__":
    unittest.main()



import tensorflow as tf


def rao_blackwell_ledoit_wolf(cov, N):
    """Rao-Blackwellized Ledoit-Wolf shrinkaged estimator of the covariance
    matrix.
    Arguments:
        cov: array, shape=(n, n), Sample covariance matrix (e.g. estimated with np.cov(X.T))
        N : int, Number of data points.
    Returns:
        Corrected covariance matrix
    [1] Chen, Yilun, Ami Wiesel, and Alfred O. Hero III. "Shrinkage
    estimation of high dimensional covariance matrices" ICASSP (2009)
    Based on: https://github.com/msmbuilder/msmbuilder/blob/master/msmbuilder/decomposition/tica.py
    """
    p = len(cov)
    assert cov.shape == (p, p)

    alpha = (N - 2) / (N * (N + 2))
    beta = ((p + 1) * N - 2) / (N * (N + 2))

    trace_cov2 = tf.reduce_sum(cov * cov)
    U = (p * trace_cov2 / tf.linalg.trace(cov) ** 2) - 1
    rho = tf.minimum(alpha + beta / U, 1)

    F = (tf.linalg.trace(cov) / p) * tf.eye(p, dtype=tf.float64)
    return (1 - rho) * cov + rho * F, rho


def calc_cov(x, y, rblw=False, use_shrinkage=False, no_normalize=False):
    """Calculates covariance matrix from a batch of tensorflow data.
    x: first set of variables as an tensorflow array
    y: second set of variables as an tensorflow array
    rblw: if to use shrinkage correction rblw
    use_shrinkage: if to use shrinkage at all
    no_normalize: do not normalize by the number of samples
    """
    x = tf.cast(x, tf.float64)
    y = tf.cast(y, tf.float64)
    N = tf.cast(tf.shape(x)[0], tf.float64)
    feat = tf.cast(tf.shape(x)[1], tf.float64)
    if not no_normalize:
        cov = 1 / (N - 1) * tf.matmul(x, y, transpose_a=True)
    else:
        cov = tf.matmul(x, y, transpose_a=True)
    if rblw and use_shrinkage:
        cov_shrink, shrinkage = rao_blackwell_ledoit_wolf(cov, N)
        return cov_shrink
    else:
        return cov


def metric_VAMP2(shift, back):
    """Returns the sum of the squared top k eigenvalues of the vamp matrix,
    with k determined by the wrapper parameter k_eig, and the vamp matrix
    defined as:
        V = cov_00 ^ -1/2 * cov_01 * cov_11 ^ -1/2
    Can be used as a metric function in model.fit()

    Arguments:
        shift: shifted data
        back: not shifted data
    Returns:
        sum of the squared k highest eigenvalues in the vamp matrix (VAMP2 score)
    """
    N = tf.shape(shift)[0]
    zt0 = back
    ztt = shift

    # shape (batch_size, output)
    zt0 = zt0 - tf.reduce_mean(zt0, axis=0, keepdims=True)
    # shape (batch_size, output)
    ztt = ztt - tf.reduce_mean(ztt, axis=0, keepdims=True)
    # Calculate the covariance matrices
    # shape (output, output)
    cov_01 = calc_cov(zt0, ztt, rblw=False, use_shrinkage=False)
    cov_10 = calc_cov(ztt, zt0, rblw=False, use_shrinkage=False)
    cov_00 = calc_cov(zt0, zt0)
    cov_11 = calc_cov(ztt, ztt)
    cov_00_inv = tf.linalg.sqrtm(tf.linalg.inv(cov_00))
    cov_11_inv = tf.linalg.sqrtm(tf.linalg.inv(cov_11))
    vamp_matrix = tf.matmul(tf.matmul(cov_00_inv, cov_01), cov_11_inv)
    # Select the all singular values of the VAMP matrix
    diag = tf.linalg.svd(vamp_matrix, compute_uv=False)
    eig_sum_sq = 1 + tf.reduce_sum(diag**2)

    return eig_sum_sq



import unittest
from easysrv import SRV
import tensorflow as tf
import os

os.environ["CUDA_DEVICE_ORDER"] = "PCI_BUS_ID"  # see issue #152
os.environ["CUDA_VISIBLE_DEVICES"] = ""


class TestSVRCreation(unittest.TestCase):
    def test_create(self):
        model = tf.keras.models.Sequential(
            [
                tf.keras.layers.Dense(
                    16,
                    kernel_regularizer=tf.keras.regularizers.L2(0.0001),
                    kernel_initializer="lecun_normal",
                ),
                tf.keras.layers.Activation("swish"),
                tf.keras.layers.BatchNormalization(),
                tf.keras.layers.Dense(
                    16,
                    kernel_regularizer=tf.keras.regularizers.L2(0.0001),
                    kernel_initializer="lecun_normal",
                ),
                tf.keras.layers.Activation("swish"),
                tf.keras.layers.BatchNormalization(),
                tf.keras.layers.Dense(
                    2, kernel_regularizer=tf.keras.regularizers.L2(0.0001)
                ),
            ]
        )
        optimizer = tf.keras.optimizers.Adam(learning_rate=0.001)
        model(tf.random.normal(shape=(20, 40)))
        srv = SRV(model=model, optimizer=optimizer, epochs=5, lagtime=1, split=0.1)


if __name__ == "__main__":
    unittest.main()



import tensorflow as tf
import numpy as np
import scipy
from easysrv.utils import calc_cov, metric_VAMP2


class SRV:
    """SRV implementation from:
    Arguments:
        model: Tensorflow Sequential Model, where the number of outputs signifies the number of slow processes.
        optimizer: Tensorflow optimizer object.
        epochs: Number of epochs used to train.
        lagtime (small tau): lagtime used for training of the SRV.
        split: split between training and validation set.
    """

    def __init__(self, model=None, optimizer=None, epochs=None, lagtime=1, split=0.1):
        self.model = model
        self.training_data = None
        self.split = split
        self.optimizer = optimizer
        self.epochs = epochs
        self.ae_lagtime = lagtime

        # eigenvectors and their stats
        self.eigenvectors_ = None
        self.eigenvalues_ = None
        self.norms_ = None
        self.means_ = None

        self.dense_out = model.layers[-1].output_shape[-1]
        if self.dense_out < 2:
            raise ValueError("Output/num of slow processes has to be 2 or more")
        # stats

        self.train_loss = []
        self.train_val_loss = []
        self.validation_vamp = []

    def fit(self, data, verbose=True):
        """Fits SRV to given data
        Arguments:
            data: A list of features extracted from trajectories as numpy arrays.
                  Also a generator that returns a numpy array will work.

        Returns:
            A dictionary with history of eigenvalues, training loss and validation loss.
        """
        split_size = int(len(data) * self.split)
        self.validation_data = data[:split_size]
        self.training_data = data[split_size:]
        self.train(verbose=verbose)
        self._calc_basis()
        history = {
            "eigenvalues": list(self.eigenvalues_.numpy()),
            "VAMP2 valid score": self.validation_vamp,
            "Training loss": self.train_loss,
            "Val Training loss": self.train_val_loss,
        }
        return history

    def transform(self, data, remove_modes=[]):
        """Transforms data onto SRV slow processes. Order is the same as in the data.
        Arguments:
            data: A list of features extracted from trajectories as numpy arrays.
                  Also a generator that returns a numpy array will work.
            remove_modes: Remove modes indexed from 0 to n, where 0 is the slowest.
                          For example you can specify [0, 2] to remove slowest and third slowest.
        Returns:
            Transformed data onto slow processes as a list of numpy arrays
        """

        all_data = []
        for datapoint in data:
            z = self.model(datapoint) - self.means_
            eigenvectors = self.eigenvectors_
            # remove some eigenvectors
            all_eig_ind = [
                i for i in range(eigenvectors.shape[0]) if i not in remove_modes
            ]
            eigv = tf.transpose(tf.gather(tf.transpose(eigenvectors), all_eig_ind))
            z_product = tf.matmul(z, eigv)
            norms = tf.transpose(tf.gather(tf.transpose(self.norms_), all_eig_ind))
            z_norm = z_product / norms
            all_data.append(z_norm.numpy())

        return all_data

    def fit_transform(self, data, verbose=True, remove_modes=[]):
        """Fits SRV to given data and transforms them onto slow processes, order is the same as in the data.
        Arguments:
            data: A list of features extracted from trajectories as numpy arrays.
                  Also a generator that returns a numpy array will work.
            remove_modes: Remove modes indexed from 0 to n, where 0 is the slowest.
                          For example you can specify [0, 2] to remove slowest and third slowest.
        Returns:
            Transformed data onto slow processes as a list of numpy arrays
        """

        self.fit(data, verbose=verbose)
        return self.transform(data, remove_modes=remove_modes)

    def train(self, verbose=True):
        """Trains SRV for a given model.
        Arguments:
            None
        Returns:
            None
        """
        for i in range(self.epochs):
            loss = []
            val_loss = []
            vamp2_met = []
            for batch in self.training_data:
                batch_shift = tf.convert_to_tensor(batch[self.ae_lagtime :])
                batch_back = tf.convert_to_tensor(batch[: -self.ae_lagtime])
                batch_ae_loss, g_norm_ae = self._train_step_vamp(
                    batch_shift,
                    batch_back,
                )
                loss.append(batch_ae_loss)
            for batch in self.validation_data:
                batch_shift = self.model(tf.convert_to_tensor(batch[self.ae_lagtime :]))
                batch_back = self.model(tf.convert_to_tensor(batch[: -self.ae_lagtime]))
                vamp2_score = metric_VAMP2(batch_back, batch_shift)
                batch_val_loss = self.loss_func_vamp(batch_back, batch_shift)
                vamp2_met.append(vamp2_score)
                val_loss.append(batch_val_loss)
            mean_loss = float(np.mean(loss))
            mean_val_loss = float(np.mean(val_loss))
            mean_vamp2 = float(np.mean(vamp2_met))
            self.train_loss.append(mean_loss)
            self.train_val_loss.append(mean_val_loss)
            self.validation_vamp.append(mean_vamp2)
            if verbose:
                print("Loss for epoch {0} is {1}".format(i, mean_loss))
                print("Val Loss for epoch {0} is {1}".format(i, mean_val_loss))
                print("Validation VAMP2 for epoch {0} is {1}".format(i, mean_vamp2))

        self._calc_basis()

    @tf.function
    def _calc_basis_estimate(self, batch_shift, batch_back):
        zt0 = self.model(batch_back)
        ztt = self.model(batch_shift)
        ztt_nom = tf.cast(ztt, tf.float64)
        zt0_nom = tf.cast(zt0, tf.float64)
        ztt = ztt_nom - tf.reduce_mean(ztt_nom, axis=0)
        zt0 = zt0_nom - tf.reduce_mean(zt0_nom, axis=0)
        # weights = self._koopman_weight(zt0, ztt)
        cov_01 = calc_cov(zt0, ztt)
        cov_10 = calc_cov(ztt, zt0)
        cov_11 = calc_cov(ztt, ztt)
        cov_00 = calc_cov(zt0, zt0)

        return cov_01, cov_10, cov_11, cov_00, zt0_nom, ztt_nom

    def _calc_basis(self):
        """Calculates basis vectors for the SRV method, based on the average
        approximated covariance matrices.
        Arguments:
            None
        Returns:
            None
        """
        M = self.dense_out
        cov_01 = tf.zeros(shape=(M, M), dtype=tf.float64)
        cov_10 = tf.zeros(shape=(M, M), dtype=tf.float64)
        cov_11 = tf.zeros(shape=(M, M), dtype=tf.float64)
        cov_00 = tf.zeros(shape=(M, M), dtype=tf.float64)
        zt0_buffer = []
        ztt_buffer = []
        obs = 0
        for batch in self.training_data:
            batch_shift = tf.convert_to_tensor(batch[self.ae_lagtime :])
            batch_back = tf.convert_to_tensor(batch[: -self.ae_lagtime])
            assert tf.rank(batch_back) == 2
            assert tf.rank(batch_shift) == 2
            (
                cov_01e,
                cov_10e,
                cov_11e,
                cov_00e,
                zt0_nom,
                ztt_nom,
            ) = self._calc_basis_estimate(batch_shift, batch_back)
            zt0_buffer.append(tf.cast(zt0_nom, tf.float32))
            ztt_buffer.append(tf.cast(ztt_nom, tf.float32))
            # weights = self._koopman_weight(zt0, ztt)
            cov_01 += cov_01e
            cov_10 += cov_10e
            cov_11 += cov_11e
            cov_00 += cov_00e
            obs += 1

        zt0_concat = tf.concat(zt0_buffer, axis=0)
        ztt_concat = tf.concat(ztt_buffer, axis=0)
        x_concat = tf.concat([ztt_concat, zt0_concat], axis=0)
        self.means_ = tf.reduce_mean(x_concat, axis=0)

        cov_01 = cov_01 / obs
        cov_10 = cov_10 / obs
        cov_11 = cov_11 / obs
        cov_00 = cov_00 / obs
        self.cov_0 = 0.5 * (cov_00 + cov_11)
        self.cov_1 = 0.5 * (cov_01 + cov_10)
        assert self.cov_0.shape[0] == zt0_nom.shape[1]
        cov_1_numpy = self.cov_1.numpy()
        cov_0_numpy = self.cov_0.numpy()
        assert cov_1_numpy.dtype == np.float64
        assert cov_0_numpy.dtype == np.float64
        eigvals, eigvecs = scipy.linalg.eigh(cov_1_numpy, b=cov_0_numpy)
        # sorts descending
        idx = np.argsort(np.abs(eigvals))[::-1]
        # remove the slowest process
        self.eigenvectors_ = tf.convert_to_tensor(eigvecs[:, idx], dtype=tf.float32)
        self.eigenvalues_ = tf.convert_to_tensor(eigvals[idx], dtype=tf.float32)
        # transform features trough eigenvectors/basis vectors
        self.means_ = tf.cast(self.means_, tf.float32)
        x_concat = tf.cast(x_concat, tf.float32)
        z = tf.matmul((x_concat - self.means_), self.eigenvectors_)
        self.norms_ = tf.math.sqrt(tf.reduce_mean(z * z, axis=0))

    @tf.function
    def loss_func_vamp(self, shift, back, reversible=False, constr_cov=True):
        """Calculates the VAMP-2 score with respect to the network lobes.

        Based on:
            https://github.com/markovmodel/deeptime/blob/master/vampnet/vampnet/vampnet.py
            https://github.com/hsidky/srv/blob/master/hde/hde.py
        Arguments:
            shift: tensorflow tensor, shifted by tau and truncated in the
                    batch dimension.

            back: tensorlfow tensor, not shifted by tau in the batch dimension and
                    truncated so by size of tau.
        Returns:
            loss_score: tensorflow tensor with shape (1, ).
        """
        N = tf.shape(shift)[0]
        ztt, zt0 = (shift, back)
        zt0 = tf.cast(zt0, tf.float64)
        ztt = tf.cast(ztt, tf.float64)
        N = tf.cast(N, tf.float64)
        # shape (batch_size, output)
        zt0_mean = tf.reduce_mean(zt0, axis=0, keepdims=True)
        ztt_mean = tf.reduce_mean(ztt, axis=0, keepdims=True)
        # Remove the mean from the data
        # shape (batch_size, output)
        zt0 = zt0 - zt0_mean
        ztt = ztt - ztt_mean
        # shape (output, output)
        cov_01 = calc_cov(zt0, ztt)
        cov_10 = calc_cov(ztt, zt0)
        cov_00 = calc_cov(zt0, zt0)
        cov_11 = calc_cov(ztt, ztt)
        cov_0 = 0.5 * (cov_00 + cov_11)
        cov_1 = 0.5 * (cov_01 + cov_10)
        L = tf.linalg.cholesky(cov_0)
        Linv = tf.linalg.inv(L)
        add_loss = 0
        A = tf.matmul(tf.matmul(Linv, cov_1), Linv, transpose_b=True)
        # make sure that all matrices are positive definitive
        lambdas, eig_v = tf.linalg.eigh(A)
        if constr_cov:
            add_loss += (
                -tf.reduce_mean(
                    tf.math.sign(tf.linalg.eigh(cov_11)[0])
                    + tf.math.sign(tf.linalg.eigh(cov_00)[0])
                )
                / 2
            )
        lambdas, eig_v = tf.linalg.eigh(A)
        # train only with respect to positive eigenvalues
        loss = -1 - tf.reduce_sum(tf.math.sign(lambdas) * lambdas**2) + add_loss
        loss = tf.cast(loss, tf.float32)

        return loss

    @tf.function
    def _train_step_vamp(self, shift, back):
        """One train step as the TF 2 graph for autoencoder.
        Arguments:
            shift: Feature matrix which is shifted by tau
            back: Feature matrix which is not shifted by tau
        Returns:
            Tuple with loss and global norm
        """
        with tf.GradientTape() as tape:
            # standarization of obs
            out_shift = self.model(shift, training=True)
            out_back = self.model(back, training=True)
            regul_loss = tf.reduce_sum(self.model.losses)
            loss = self.loss_func_vamp(out_shift, out_back)
            loss += regul_loss
            trainable_var = self.model.trainable_variables
        grads = tape.gradient((loss), trainable_var)
        global_norm = tf.linalg.global_norm(grads)
        self.optimizer.apply_gradients(zip(grads, trainable_var))

        return loss, global_norm



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="easysrv",
    author="Daniel Wiczew",
    author_email="daniel.wiczew@gmail.com",
    description="Easy to use State Reversible Vampnet with fit, transform and fittransform methods",
    keywords="machine learning, molecular dynamics, tensorflow",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/DanielWicz/easysrv",
    project_urls={
        "Documentation": "https://github.com/DanielWicz/easysrv",
        "Bug Reports": "https://github.com/DanielWicz/easysrv/issues",
        "Source Code": "https://github.com/DanielWicz/easysrv",
        # 'Funding': '',
        # 'Say Thanks!': '',
    },
    package_dir={"": "src"},
    packages=setuptools.find_packages(where="src"),
    classifiers=[
        # see https://pypi.org/classifiers/
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3 :: Only",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.6",
    install_requires=["numpy", "tensorflow", "scipy"],
    extras_require={
        "dev": ["check-manifest"],
    },
)
